@@ -17,6 +17,7 @@
 #
 
 from __future__ import print_function
+from datetime import date, timedelta
 import sys
 import os
 import signal
@@ -24,6 +25,7 @@ import getopt
 import json
 import pprint
 import logging
+import ssl
 try:
     from urllib.parse import urlparse
 except ImportError:
@@ -63,15 +65,20 @@ def main():
         handler.setFormatter(formatter)
         logger.addHandler(handler)
 
+    if options['cafile'] or options['capath'] or options['ssl']:
+        ssl_context = create_ssl_context(options['cafile'],
+                                         options['capath'],
+                                         options['ssl'])
+    elif options['ssl'] is None:
+        ssl_context = pan.wfapi.cloud_ssl_context()
+
     try:
         wfapi = pan.wfapi.PanWFapi(tag=options['tag'],
                                    api_key=options['api_key'],
                                    hostname=options['hostname'],
                                    timeout=options['timeout'],
                                    http=options['http'],
-                                   cacloud=options['cacloud'],
-                                   cafile=options['cafile'],
-                                   capath=options['capath'])
+                                   ssl_context=ssl_context)
 
     except pan.wfapi.PanWFapiError as msg:
         print('pan.wfapi.PanWFapi:', msg, file=sys.stderr)
@@ -112,6 +119,34 @@ def main():
             print_status(wfapi, action)
             print_response(wfapi, options)
 
+        if options['submit-link'] is not None:
+            action = 'submit'
+            kwargs = {}
+            kwargs['links'] = process_arg(options['submit-link'], list=True)
+
+            wfapi.submit(**kwargs)
+            print_status(wfapi, action)
+            print_response(wfapi, options)
+
+        if options['change-request']:
+            action = 'change-request'
+            kwargs = {}
+            if len(hashes) > 1:
+                print('Only 1 hash allowed for %s' % action, file=sys.stderr)
+                sys.exit(1)
+            if len(hashes) == 1:
+                kwargs['hash'] = hashes[0]
+            if options['new-verdict'] is not None:
+                kwargs['verdict'] = process_verdict(options['new-verdict'])
+            if options['email'] is not None:
+                kwargs['email'] = options['email']
+            if options['comment'] is not None:
+                kwargs['comment'] = process_arg(options['comment'])
+
+            wfapi.change_request(**kwargs)
+            print_status(wfapi, action)
+            print_response(wfapi, options)
+
         if options['report']:
             action = 'report'
             kwargs = {}
@@ -181,6 +216,18 @@ def main():
             kwargs = {}
             if options['date'] is not None:
                 kwargs['date'] = options['date']
+                try:
+                    x = int(options['date'])
+                except ValueError:
+                    pass
+                else:
+                    if x < 1:
+                        d = date.today()
+                        d = d - timedelta(-x)
+                        kwargs['date'] = d.isoformat()
+                        if options['debug']:
+                            print('relative date(%d): %s' % (x, kwargs['date']),
+                                  file=sys.stderr)
 
             wfapi.verdicts_changed(**kwargs)
             print_status(wfapi, action)
@@ -239,9 +286,51 @@ def validate_hash(hash):
         sys.exit(1)
 
 
+def process_arg(s, list=False):
+    stdin_char = '-'
+
+    if s == stdin_char:
+        lines = sys.stdin.readlines()
+    else:
+        try:
+            f = open(s)
+            lines = f.readlines()
+            f.close()
+        except IOError:
+            lines = [s]
+
+    if debug > 1:
+        print('lines:', lines, file=sys.stderr)
+
+    if list:
+        l = [x.rstrip('\r\n') for x in lines]
+        return l
+
+    lines = ''.join(lines)
+    return lines
+
+
+def process_verdict(verdict):
+    verdicts = {
+        'benign': pan.wfapi.BENIGN,
+        'malware': pan.wfapi.MALWARE,
+        'grayware': pan.wfapi.GRAYWARE,
+    }
+
+    try:
+        int(verdict)
+        return verdict
+    except ValueError:
+        if verdict in verdicts:
+            return str(verdicts[verdict])
+        return verdict
+
+
 def parse_opts():
     options = {
         'submit': None,
+        'submit-link': None,
+        'change-request': False,
         'report': False,
         'verdict': False,
         'sample': False,
@@ -249,6 +338,9 @@ def parse_opts():
         'changed': False,
         'hash': [],
         'platform': None,
+        'new-verdict': None,
+        'email': None,
+        'comment': None,
         'testfile': False,
         'format': None,
         'date': None,
@@ -256,7 +348,7 @@ def parse_opts():
         'api_key': None,
         'hostname': None,
         'http': False,
-        'cacloud': True,
+        'ssl': None,
         'cafile': None,
         'capath': None,
         'print_xml': False,
@@ -270,11 +362,13 @@ def parse_opts():
 
     short_options = 'K:h:xpjHDt:T:'
     long_options = ['version', 'help',
-                    'submit=', 'report', 'verdict', 'sample',
+                    'submit=', 'submit-link=',
+                    'change-request', 'report', 'verdict', 'sample',
                     'pcap', 'changed',
                     'hash=', 'platform=', 'testfile',
+                    'new-verdict=', 'email=', 'comment=',
                     'format=', 'date=', 'dst=',
-                    'http', 'nocacloud', 'cafile=', 'capath=',
+                    'http', 'ssl=', 'cafile=', 'capath=',
                     ]
 
     try:
@@ -290,6 +384,10 @@ def parse_opts():
             pass
         elif opt == '--submit':
             options['submit'] = arg
+        elif opt == '--submit-link':
+            options['submit-link'] = arg
+        elif opt == '--change-request':
+            options['change-request'] = True
         elif opt == '--report':
             options['report'] = True
         elif opt == '--verdict':
@@ -304,6 +402,12 @@ def parse_opts():
             options['hash'].append(arg)
         elif opt == '--platform':
             options['platform'] = arg
+        elif opt == '--new-verdict':
+            options['new-verdict'] = arg
+        elif opt == '--email':
+            options['email'] = arg
+        elif opt == '--comment':
+            options['comment'] = arg
         elif opt == '--testfile':
             options['testfile'] = True
         elif opt == '--format':
@@ -318,8 +422,12 @@ def parse_opts():
             options['hostname'] = arg
         elif opt == '--http':
             options['http'] = True
-        elif opt == '--nocacloud':
-            options['cacloud'] = False
+        elif opt == '--ssl':
+            if arg in ['default', 'noverify', 'cacloud']:
+                options['ssl'] = arg
+            else:
+                print('Invalid --ssl option:', arg)
+                sys.exit(1)
         elif opt == '--cafile':
             options['cafile'] = arg
         elif opt == '--capath':
@@ -360,6 +468,36 @@ def parse_opts():
     return options
 
 
+def create_ssl_context(cafile, capath, ssl_option):
+    # PEP 0476
+    if (sys.version_info.major == 2 and sys.hexversion >= 0x02070900 or
+            sys.version_info.major == 3 and sys.hexversion >= 0x03040300):
+        if cafile or capath:
+            try:
+                ssl_context = ssl.create_default_context(
+                    purpose=ssl.Purpose.SERVER_AUTH,
+                    cafile=cafile,
+                    capath=capath)
+            except Exception as e:
+                print('cafile or capath invalid: %s' % e, file=sys.stderr)
+                sys.exit(1)
+        elif ssl_option:
+            if ssl_option == 'cacloud':
+                ssl_context = pan.wfapi.cloud_ssl_context()
+            elif ssl_option == 'noverify':
+                ssl_context = ssl._create_unverified_context()
+            elif ssl_option == 'default':
+                ssl_context = None
+
+        return ssl_context
+
+    print('Warning: Python %d.%d.%d: cafile, capath and ssl ignored' %
+          (sys.version_info.major, sys.version_info.minor,
+           sys.version_info.micro), file=sys.stderr)
+
+    return None
+
+
 def print_status(wfapi, action, exception_msg=None):
     print(action, end='', file=sys.stderr)
 
@@ -480,6 +618,8 @@ def set_encoding():
 def usage():
     usage = '''%s [options]
     --submit path|url     submit file or URL to WildFire for analysis
+    --submit-link link    submit links to WildFire for analysis
+    --change-request      request review of sample's verdict
     --report              get WildFire report
     --verdict             get WildFire sample verdict
     --sample              get WildFire sample file
@@ -487,9 +627,13 @@ def usage():
     --changed             get changed verdicts
     --hash hash           query MD5 or SHA256 hash
     --platform id         platform ID for sandbox environment
+    --new-verdict verdict benign|malware|grayware
+    --email address       notification e-mail address
+    --comment comment     change request explanation
     --testfile            get sample malware test file
     --format format       report output format
-    --date date           start date for changed verdicts (YYYY-MM-DD)
+    --date date           start date for changed verdicts
+                          (YYYY-MM-DD or -days)
     --dst dst             save file to directory or path
     -K api_key            WildFire API key
     -h hostname           WildFire hostname
@@ -500,7 +644,7 @@ def usage():
     -t tag                .panrc tagname
     -T seconds            urlopen() timeout
     --http                use http URL scheme (default https)
-    --nocacloud           disable default cloud CA certificate verification
+    --ssl opt             SSL verify option: default|noverify|cacloud
     --cafile path         file containing CA certificates
     --capath path         directory of hashed certificate files
     --version             display version
