@@ -14,10 +14,11 @@
 # OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 #
 
-"""Interface to the WildFire  API
+"""Interface to the WildFire API
 
 The pan.wfapi module implements the PanWFapi class.  It provides an
-interface to the WildFire API on Palo Alto Networks' WildFire Cloud.
+interface to the WildFire API on Palo Alto Networks' WildFire Cloud
+and WildFire appliance.
 """
 
 # XXX Using the requests module which uses urllib3 and has support
@@ -27,26 +28,25 @@ interface to the WildFire API on Palo Alto Networks' WildFire Cloud.
 # to be revisited as some parts of this are not clean.
 
 from __future__ import print_function
+import socket
 import sys
 import re
 import os
-from tempfile import NamedTemporaryFile
 from io import BytesIO
 import email.utils
 import logging
 try:
     # 3.2
-    from urllib.request import Request, urlopen, \
-        build_opener, install_opener, HTTPErrorProcessor
-    from urllib.request import HTTPSHandler, OpenerDirector
+    from urllib.request import Request, \
+        build_opener, HTTPErrorProcessor, HTTPSHandler
     from urllib.error import URLError
     from urllib.parse import urlencode
     from http.client import responses
     _legacy_urllib = False
 except ImportError:
     # 2.7
-    from urllib2 import Request, urlopen, URLError, \
-        build_opener, install_opener, HTTPErrorProcessor
+    from urllib2 import Request, URLError, \
+        build_opener, HTTPErrorProcessor, HTTPSHandler
     from urllib import urlencode
     from httplib import responses
     _legacy_urllib = True
@@ -55,19 +55,11 @@ import xml.etree.ElementTree as etree
 from . import __version__, DEBUG1, DEBUG2, DEBUG3
 import pan.rc
 
-import socket
 try:
     import ssl
 except ImportError:
     raise ValueError('SSL support not available')
 
-try:
-    ssl.SSLContext(ssl.PROTOCOL_SSLv23)
-except AttributeError:
-    _have_sslcontext = False
-else:
-    _have_sslcontext = True
-
 _cloud_server = 'wildfire.paloaltonetworks.com'
 _encoding = 'utf-8'
 _rfc2231_encode = False
@@ -117,13 +109,7 @@ def _isbytes(s):
 
 
 class PanWFapiError(Exception):
-    def __init__(self, msg):
-        self.msg = msg
-
-    def __str__(self):
-        if self.msg is None:
-            return ''
-        return self.msg
+    pass
 
 
 class PanWFapi:
@@ -133,29 +119,18 @@ class PanWFapi:
                  api_key=None,
                  timeout=None,
                  http=False,
-                 cacloud=True,
-                 cafile=None,
-                 capath=None):
+                 ssl_context=None):
         self._log = logging.getLogger(__name__).log
         self.tag = tag
         self.hostname = hostname
         self.api_key = None
         self.timeout = timeout
-        self.cafile = cafile
-        self.capath = capath
+        self.ssl_context = ssl_context
 
         self._log(DEBUG3, 'Python version: %s', sys.version)
         self._log(DEBUG3, 'xml.etree.ElementTree version: %s', etree.VERSION)
         self._log(DEBUG3, 'pan-python version: %s', __version__)
 
-        if ((cacloud and sys.hexversion >= 0x03020000) and
-                (self.cafile is None and self.capath is None)):
-            tempfile = self.__cacloud()
-            if tempfile is None:
-                raise PanWFapiError(self._msg)
-            self.cacloud_tempfile = tempfile
-            self.cafile = self.cacloud_tempfile.name
-
         if self.timeout is not None:
             try:
                 self.timeout = int(self.timeout)
@@ -164,6 +139,12 @@ class PanWFapi:
             except ValueError:
                 raise PanWFapiError('Invalid timeout: %s' % self.timeout)
 
+        if ssl_context is not None:
+            try:
+                ssl.SSLContext(ssl.PROTOCOL_SSLv23)
+            except AttributeError:
+                raise PanXapiError('SSL module has no SSLContext()')
+
         init_panrc = {}  # .panrc args from constructor
         if hostname is not None:
             init_panrc['hostname'] = hostname
@@ -343,32 +324,6 @@ class PanWFapi:
         self._log(DEBUG3, 'xml_root.decode(): %s', type(s.decode(_encoding)))
         return s.decode(_encoding)
 
-    # see http://bugs.python.org/issue18543
-    # this is a modified urllib.request.urlopen()
-    @staticmethod
-    def _urlopen(url, data=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
-                 cafile=None, capath=None, cadefault=False):
-
-        def http_response(request, response):
-            return response
-
-        http_error_processor = HTTPErrorProcessor()
-        http_error_processor.https_response = http_response
-
-        if _have_sslcontext and (cafile or capath or cadefault):
-            context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
-            context.options |= ssl.OP_NO_SSLv2
-            context.verify_mode = ssl.CERT_REQUIRED
-            if cafile or capath:
-                context.load_verify_locations(cafile, capath)
-            else:
-                context.set_default_verify_paths()
-            https_handler = HTTPSHandler(context=context, check_hostname=True)
-            opener = build_opener(https_handler, http_error_processor)
-        else:
-            opener = build_opener(http_error_processor)
-        return opener.open(url, data, timeout)
-
 # XXX Unicode notes
 # 2.7
 # decode() str (bytes) -> unicode
@@ -401,23 +356,15 @@ class PanWFapi:
         kwargs = {
             'url': request,
             }
-        # Changed in version 3.2: cafile and capath were added.
-        if sys.hexversion >= 0x03020000:
-            kwargs['cafile'] = self.cafile
-            kwargs['capath'] = self.capath
-        # Changed in version 3.3: cadefault added
-        if sys.hexversion >= 0x03030000:
-            pass
-#            kwargs['cadefault'] = True
+
+        if self.ssl_context is not None:
+            kwargs['context'] = self.ssl_context
 
         if self.timeout is not None:
             kwargs['timeout'] = self.timeout
 
         try:
-
-            response = PanWFapi._urlopen(**kwargs)
-
-        # invalid cafile, capath
+            response = self._urlopen(**kwargs)
         except (URLError, IOError) as e:
             self._log(DEBUG2, 'urlopen() exception: %s', sys.exc_info())
             self._msg = str(e)
@@ -614,21 +561,25 @@ class PanWFapi:
 
     def submit(self,
                file=None,
-               url=None):
+               url=None,
+               links=None):
         self.__clear_response()
 
-        if (file is not None and url is not None):
-            raise PanWFapiError('must submit file or url, not both')
+        if (sum(bool(x) for x in [file, url, links]) != 1):
+            raise PanWFapiError('must submit one of file, url or links')
 
         if file is not None:
             request_uri = '/publicapi/submit/file'
         elif url is not None:
             request_uri = '/publicapi/submit/url'
-        else:
-            raise PanWFapiError('file or url not specified')
+        elif len(links) < 2:
+            request_uri = '/publicapi/submit/link'
+        elif len(links) > 1:
+            request_uri = '/publicapi/submit/links'
 
         form = _MultiPartFormData()
         form.add_field('apikey', self.api_key)
+
         if file is not None:
             buf = self._read_file(file)
             if buf is None:
@@ -639,6 +590,19 @@ class PanWFapi:
         if url is not None:
             form.add_field('url', url)
 
+        if links is not None:
+            if len(links) == 1:
+                form.add_field('link', links[0])
+            elif len(links) > 1:
+                magic = 'panlnk'  # XXX should be optional in future
+                # XXX requires filename in Content-Disposition header
+                if links[0] == magic:
+                    form.add_file(filename='pan',
+                                  body='\n'.join(links))
+                else:
+                    form.add_file(filename='pan',
+                                  body=magic + '\n' + '\n'.join(links))
+
         headers = form.http_headers()
         body = form.http_body()
 
@@ -650,15 +614,70 @@ class PanWFapi:
         if not self.__set_response(response):
             raise PanWFapiError(self._msg)
 
-    def __cacloud(self):
-        # WildFire cloud cafile:
-        #   https://certs.godaddy.com/anonymous/repository.pki
-        #   Go Daddy Class 2 Certification Authority Root Certificate
-        # use:
-        #   $ openssl x509 -in wfapi.py -text
-        # to view text form.
+    def change_request(self,
+                       hash=None,
+                       verdict=None,
+                       email=None,
+                       comment=None):
+        self.__clear_response()
 
-        gd_class2_root_crt = b'''
+        request_uri = '/publicapi/submit/change-request'
+
+        form = _MultiPartFormData()
+        form.add_field('apikey', self.api_key)
+        if hash is not None:
+            form.add_field('hash', hash)
+        if verdict is not None:
+            form.add_field('verdict', verdict)
+        if email is not None:
+            form.add_field('email', email)
+        if comment is not None:
+            form.add_field('comment', comment)
+
+        headers = form.http_headers()
+        body = form.http_body()
+
+        response = self.__api_request(request_uri=request_uri,
+                                      body=body, headers=headers)
+        if not response:
+            raise PanWFapiError(self._msg)
+
+        if not self.__set_response(response):
+            raise PanWFapiError(self._msg)
+
+    # allow non-2XX error codes
+    # see http://bugs.python.org/issue18543 for why we can't just
+    # install a new HTTPErrorProcessor()
+    @staticmethod
+    def _urlopen(url, data=None,
+                 timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
+                 cafile=None, capath=None, cadefault=False,
+                 context=None):
+
+        def http_response(request, response):
+            return response
+
+        http_error_processor = HTTPErrorProcessor()
+        http_error_processor.https_response = http_response
+
+        if context:
+            https_handler = HTTPSHandler(context=context)
+            opener = build_opener(https_handler, http_error_processor)
+        else:
+            opener = build_opener(http_error_processor)
+
+        return opener.open(url, data, timeout)
+
+
+def cloud_ssl_context():
+    # WildFire cloud cafile:
+    #   https://certs.godaddy.com/anonymous/repository.pki
+    #   Go Daddy Class 2 Certification Authority Root Certificate
+    # use:
+    #   $ openssl x509 -in wfapi.py -text
+    # to view text form.
+
+    gd_class2_root_crt = b'''
 -----BEGIN CERTIFICATE-----
 MIIEADCCAuigAwIBAgIBADANBgkqhkiG9w0BAQUFADBjMQswCQYDVQQGEwJVUzEh
 MB8GA1UEChMYVGhlIEdvIERhZGR5IEdyb3VwLCBJbmMuMTEwLwYDVQQLEyhHbyBE
@@ -685,18 +704,15 @@ ReYNnyicsbkqWletNw+vHX/bvZ8=
 -----END CERTIFICATE-----
 '''
 
-        try:
-            tf = NamedTemporaryFile(suffix='.crt')
-            tf.write(gd_class2_root_crt)
-            tf.flush()
-        except (OSError, IOError) as e:
-            self._msg = "Can't create cloud cafile: %s" % e
-            return None
-
-        self._log(DEBUG2, '__cacloud: %s', tf.name)
-
-        return tf
-
+    if (sys.version_info.major == 2 and sys.hexversion >= 0x02070900 or
+            sys.version_info.major == 3 and sys.hexversion >= 0x03040300):
+        # XXX python >= 2.7.9 needs catada as Unicode, or we get:
+        # 'ssl.SSLError: nested asn1 error'
+        return ssl.create_default_context(
+            purpose=ssl.Purpose.SERVER_AUTH,
+            cadata=gd_class2_root_crt.decode())
+    else:
+        return None
 
 # Minimal RFC 2388 implementation
 
@@ -712,6 +728,7 @@ ReYNnyicsbkqWletNw+vHX/bvZ8=
 # XXXfilecontents
 # --___XXX--
 
+
 class _MultiPartFormData:
     def __init__(self):
         self._log = logging.getLogger(__name__).log
